import copy

from databricks.feature_store.constants import MAX_PRIMARY_KEY_STRING_LENGTH_BYTES
from databricks.feature_store.publish_engine.publish_sql_engine import (
    PublishSqlEngine,
    get_string_primary_keys,
    set_primary_keys_not_nullable,
)
from databricks.feature_store.utils.spark_utils import (
    get_columns_of_type_short,
    serialize_complex_data_types,
)


class PublishSqlServerEngine(PublishSqlEngine):
    """
    SqlServerEngine supports SQL Server 2019 and newer due to its reliance on UTF-8 collations.
    (https://docs.microsoft.com/en-us/sql/relational-databases/collations/collation-and-unicode-support?view=sql-server-2017#utf8)
    """

    DEFAULT_DRIVER = "com.microsoft.sqlserver.jdbc.SQLServerDriver"
    COLLATION = "Latin1_General_100_BIN2_UTF8"
    MAX_NUM_STRING_PK_COLUMNS = 2
    PRIMARY_KEY_STRING_ERROR = "String or binary data would be truncated"

    def __init__(self, online_store, spark_session):
        super().__init__(online_store, spark_session)

    def get_online_tables(self, table_names=[]):
        query = (
            f"SELECT {PublishSqlEngine.TABLE_NAME} "
            f"FROM {PublishSqlEngine.INFORMATION_SCHEMA}.{PublishSqlEngine.TABLES} "
            f"WHERE {PublishSqlEngine.TABLE_CATALOG}='{self.online_store.database_name}'"
        )
        if len(table_names) > 0:
            table_names_str = ", ".join(
                [f"'{table_name}'" for table_name in table_names]
            )
            query += f" AND {PublishSqlEngine.TABLE_NAME} IN ({table_names_str})"
        return self._run_sql_query(query)

    def get_column_types(self, table_name):
        query = (
            f"SELECT {PublishSqlEngine.COLUMN_NAME}, {PublishSqlEngine.DATA_TYPE}, {PublishSqlEngine.NUMERIC_PRECISION}, {PublishSqlEngine.NUMERIC_SCALE} "
            f"FROM {PublishSqlEngine.INFORMATION_SCHEMA}.{PublishSqlEngine.COLUMNS} "
            f"WHERE {PublishSqlEngine.TABLE_CATALOG}='{self.online_store.database_name}' AND {PublishSqlEngine.TABLE_NAME}='{table_name}'"
        )
        return self._run_sql_query(query)

    def add_columns(self, table_name, columns):
        if len(columns) == 0:
            return
        new_columns_str = ", ".join(
            [f"{self._sql_safe_name(name)} {type}" for (name, type) in columns]
        )
        self._run_sql_update(
            f"ALTER TABLE {self._sql_safe_name(table_name)} ADD {new_columns_str}"
        )

    def create_table_like(self, table_name, like_table_name, primary_keys):
        with self.in_transaction():
            primary_keys_str = ", ".join(
                f"{self._sql_safe_name(pk)}" for pk in primary_keys
            )
            update_cmd = (
                "IF NOT EXISTS "
                f"(SELECT * FROM {PublishSqlEngine.INFORMATION_SCHEMA}.{PublishSqlEngine.TABLES} WHERE {PublishSqlEngine.TABLE_NAME} = '{table_name}') "
                "BEGIN "
                f"SELECT * INTO {self._sql_safe_name(table_name)} FROM {self._sql_safe_name(like_table_name)} WHERE 0 = 1; "
                f"ALTER TABLE {self._sql_safe_name(table_name)} ADD PRIMARY KEY({primary_keys_str}); "
                "END"
            )
            return self._run_sql_update(update_cmd)

    def rename_table(self, from_name, to_name):
        self._run_sql_update(
            f"EXEC sp_rename {self._sql_safe_name(from_name)}, {self._sql_safe_name(to_name)}"
        )

    def merge_table_into(self, dst_table_name, src_table_name, columns, primary_keys):
        columns_str = ", ".join(f"{self._sql_safe_name(c)}" for c in columns)
        update_columns = [col for col in columns if col not in primary_keys]
        update_str = ", ".join(
            [
                f"{self._sql_safe_name(feat)}={self._sql_safe_name(src_table_name)}.{self._sql_safe_name(feat)}"
                for feat in update_columns
            ]
        )
        primary_keys_str = " AND ".join(
            [
                f"{self._sql_safe_name(dst_table_name)}.{self._sql_safe_name(key)}={self._sql_safe_name(src_table_name)}.{self._sql_safe_name(key)}"
                for key in primary_keys
            ]
        )
        values_str = ", ".join(
            [
                f"{self._sql_safe_name(src_table_name)}.{self._sql_safe_name(col)}"
                for col in columns
            ]
        )

        update_cmd = (
            f"MERGE {self._sql_safe_name(dst_table_name)} "
            f"USING {self._sql_safe_name(src_table_name)} ON {primary_keys_str} "
            "WHEN MATCHED THEN "
            f"UPDATE SET {update_str} "
            "WHEN NOT MATCHED BY TARGET THEN "
            f"INSERT ({columns_str}) VALUES ({values_str})"
        )
        self._run_sql_update(update_cmd)

    def _alter_string_pk_collation(self, table_name, primary_key):
        self._run_sql_update(
            f"ALTER TABLE {self._sql_safe_name(table_name)} "
            f"ALTER COLUMN {self._sql_safe_name(primary_key)} VARCHAR({MAX_PRIMARY_KEY_STRING_LENGTH_BYTES}) "
            f"COLLATE {self.COLLATION} "
            f"NOT NULL"
        )

    def _update_create_table_column_types_option(self, writer, short_cols):
        # Explicitly specify the SHORT type to prevent upcasting to INTEGER by the Spark writer
        # SHORT is specified rather than Sql Server's SMALLINT as the documentation asks for a Spark Sql type
        # However, SHORT and SMALLINT are equivalent. SHORT will be written as SMALLINT in Sql Server
        # https://spark.apache.org/docs/latest/sql-data-sources-jdbc.html
        if len(short_cols) > 0:
            short_col_types = [f"{col} SHORT" for col in short_cols]
            writer = writer.option(
                "createTableColumnTypes",
                ", ".join(short_col_types),
            )
        return writer

    def create_empty_table(
        self, table_name, schema, jdbc_url, connection_properties, primary_keys
    ):
        # MSSQL requires columns be non-nullable before setting them as
        # primary keys. We set the primary key fields in the schema
        # to be non-nullable so they are defined as not-nullable
        # in the online store schema.
        schema = copy.deepcopy(schema)
        set_primary_keys_not_nullable(schema, primary_keys)
        empty_df = self._make_empty_df(schema)

        string_pk_cols = get_string_primary_keys(schema, primary_keys)
        short_cols = get_columns_of_type_short(schema)

        # SQL Server 2019 does not error out if the primary key maximum length (900 bytes) is exceeded.
        # Therefore, we explicitly guard it from feature tables that have more than two
        # 400-byte string columns in the primary key. Exceeding the maximum length is still
        # possible using additional column types but is unlikely.
        if len(string_pk_cols) > self.MAX_NUM_STRING_PK_COLUMNS:
            raise ValueError(
                f"Feature tables with more than {self.MAX_NUM_STRING_PK_COLUMNS} string columns "
                f"in their primary key cannot be published to SQL Server; "
                f"found {len(string_pk_cols)} string columns in the primary key."
            )

        serialized_df = serialize_complex_data_types(empty_df)
        writer = self._update_create_table_column_types_option(
            serialized_df.write, short_cols
        )
        writer.jdbc(
            url=jdbc_url,
            table=table_name,
            mode="overwrite",
            properties=connection_properties,
        )
        for string_pk in string_pk_cols:
            self._alter_string_pk_collation(table_name, string_pk)

        self.add_primary_keys(table_name, primary_keys)

    @classmethod
    def _sql_safe_name(cls, name):
        # MSSQL requires [xxx] format to safely handle identifiers that contain special characters or are reserved words.
        return f"[{name}]"

    @property
    def jdbc_url(self):
        return f"jdbc:sqlserver://{self.online_store.hostname}:{self.online_store.port};database={self.online_store.database_name}"

    @property
    def jdbc_properties(self):
        return {
            "user": self.online_store._lookup_jdbc_auth_user_with_write_permissions(),
            "password": self.online_store._lookup_password_with_write_permissions(),
            "driver": self.online_store.driver or self.DEFAULT_DRIVER,
        }



from mlflow.utils import databricks_utils
import logging
import pkg_resources
import yaml
from typing import List

_logger = logging.getLogger(__name__)

# Load in the list of valid default header keys from the headers.yaml bundled resource
constants_yml = pkg_resources.resource_string(__name__, "headers.yaml")
if constants_yml is None or len(constants_yml) == 0:
    raise Exception(
        "Missing headers.yaml package resource.  This indicates a packaging error."
    )
constants = yaml.safe_load(constants_yml)
valid_header_keys = {}
for key in constants.keys():
    valid_header_keys[key] = set(constants.get(key, []))

# Default header key for the feature store client API method that originated this request context
FEATURE_STORE_METHOD_NAME = "feature-store-method-name"

# Other default header keys
CLUSTER_ID = "cluster_id"
NOTEBOOK_ID = "notebook_id"
JOB_ID = "job_id"
JOB_RUN_ID = "job_run_id"
JOB_TYPE = "job_type"

# custom header keys
PUBLISH_AUTH_TYPE = "publish_auth_type"

DEFAULT_HEADERS = "default_headers"
# Valid header keys and header values for FEATURE_STORE_METHOD_NAME header key.
GET_TABLE = "get_table"
GET_FEATURE_TABLE = "get_feature_table"
CREATE_TABLE = "create_table"
CREATE_FEATURE_TABLE = "create_feature_table"
REGISTER_TABLE = "register_table"
PUBLISH_TABLE = "publish_table"
WRITE_TABLE = "write_table"
READ_TABLE = "read_table"
DROP_TABLE = "drop_table"
LOG_MODEL = "log_model"
SCORE_BATCH = "score_batch"
CREATE_TRAINING_SET = "create_training_set"
GET_MODEL_SERVING_METADATA = "get_model_serving_metadata"
SET_FEATURE_TABLE_TAG = "set_feature_table_tag"
DELETE_FEATURE_TABLE_TAG = "delete_feature_table_tag"
ADD_DATA_SOURCES = "add_data_sources"
DELETE_DATA_SOURCES = "delete_data_sources"
TEST_ONLY_METHOD = "test_only_method"


class RequestContext:

    """
    An object for instrumenting the feature store client usage patterns.  Client methods in the public
    API should create a RequestContext and pass it down the callstack to the catalog client, which will
    add all relevant context to the outgoing requests as HTTP headers.  The catalog service will read the
    headers and record in usage logs.
    """

    # The list of valid header values for the FEATURE_STORE_METHOD_NAME header key
    valid_feature_store_method_names = [
        GET_TABLE,
        GET_FEATURE_TABLE,
        CREATE_TABLE,
        CREATE_FEATURE_TABLE,
        PUBLISH_TABLE,
        WRITE_TABLE,
        READ_TABLE,
        REGISTER_TABLE,
        DROP_TABLE,
        LOG_MODEL,
        SCORE_BATCH,
        CREATE_TRAINING_SET,
        GET_MODEL_SERVING_METADATA,
        SET_FEATURE_TABLE_TAG,
        DELETE_FEATURE_TABLE_TAG,
        ADD_DATA_SOURCES,
        DELETE_DATA_SOURCES,
        TEST_ONLY_METHOD,
    ]

    def __init__(self, feature_store_method_name: str, custom_headers: dict = {}):
        """
        Initializer

        :param feature_store_method_name: The feature store method creating this request context.
        """
        if (
            feature_store_method_name
            not in RequestContext.valid_feature_store_method_names
        ):
            raise ValueError(
                f"Invalid feature store method name given: {feature_store_method_name}"
            )

        # Create default headers
        default_headers = self._create_default_headers(feature_store_method_name)

        # Ensure that no header keys outside of those declared in headers.yaml have snuck into the codebase
        self._validate_headers(
            feature_store_method_name,
            default_headers,
            custom_headers,
            valid_header_keys,
        )

        # Save default in internal headers map
        self._headers = default_headers
        # Save custom headers
        self._headers.update(custom_headers)

    def __eq__(self, other):
        """
        Override equality testing to compare the internal state rather than comparing by reference.
        Curently only needed for testing purposes.  If additional state variables are added to this
        object this method will need to be updated accordingly.
        """
        if not isinstance(other, RequestContext):
            return NotImplemented
        return self.__dict__ == other.__dict__

    def _create_default_headers(self, feature_store_method_name):
        """
        Create the default headers that will be sent with every RPC request from the client.
        """
        default_headers = {FEATURE_STORE_METHOD_NAME: feature_store_method_name}
        try:
            if databricks_utils.is_in_cluster():
                default_headers[CLUSTER_ID] = databricks_utils.get_cluster_id()
            if databricks_utils.is_in_databricks_notebook():
                default_headers[NOTEBOOK_ID] = databricks_utils.get_notebook_id()
            if databricks_utils.is_in_databricks_job():
                default_headers[JOB_ID] = databricks_utils.get_job_id()
                default_headers[JOB_RUN_ID] = databricks_utils.get_job_run_id()
                default_headers[JOB_TYPE] = databricks_utils.get_job_type()
        except Exception as e:
            _logger.warning(
                "Exeption while adding standard headers, some headers will not be added.",
                exc_info=e,
            )
        return default_headers

    def _validate_headers(
        self,
        feature_store_method_name,
        default_headers,
        custom_headers,
        valid_header_keys,
    ):
        """
        Ensure that all headers are in the list of valid headers expected by the catalog service.
        This prevents any headers being added to the client while forgetting to add to the
        catalog service, since both share the headers.yaml file.
        """
        unknown_header_keys = []
        for key in default_headers.keys():
            if key not in valid_header_keys[DEFAULT_HEADERS]:
                unknown_header_keys.append(key)
        for key in custom_headers.keys():
            if (feature_store_method_name not in valid_header_keys) or (
                key not in valid_header_keys[feature_store_method_name]
            ):
                unknown_header_keys.append(key)
        if len(unknown_header_keys) > 0:
            raise Exception(
                f'Unknown header key{"s" if len(unknown_header_keys) > 1 else ""}: '
                f'{", ".join(unknown_header_keys)}. Please add to headers.yaml'
            )

    def get_header(self, header_name: str):
        """
        Get the specified header, or return None if there is no corresponding header.
        """
        return self._headers.get(header_name)

    def get_headers(self) -> map:
        """
        Get the stored headers.
        """
        return dict.copy(self._headers)



import logging
from typing import List, Union, Dict, Any, Optional

import databricks.feature_store.utils.validation_utils
from databricks.feature_store.constants import (
    OVERWRITE,
    MERGE,
    DATA_TYPES_REQUIRES_DETAILS,
    _DEFAULT_WRITE_STREAM_TRIGGER,
    _WARN,
    _ERROR,
    _SOURCE_FORMAT_DELTA,
)
from databricks.feature_store.api.proto.feature_catalog_pb2 import ProducerAction
from databricks.feature_store.entities.data_type import DataType
from databricks.feature_store.entities.store_type import StoreType
from databricks.feature_store.entities.feature import Feature
from databricks.feature_store.entities.feature_table import FeatureTable

from databricks.feature_store.entities.key_spec import KeySpec
from databricks.feature_store.hive_client import HiveClient
from databricks.feature_store.catalog_client import CatalogClient
from databricks.feature_store.utils import utils
from databricks.feature_store.utils.spark_listener import SparkSourceListener
from databricks.feature_store.utils.sql import is_qualified_table_name
from databricks.feature_store.utils import request_context
from databricks.feature_store.utils.request_context import (
    RequestContext,
)
from databricks.feature_store.utils import schema_utils
from databricks.feature_store._hive_client_helper import HiveClientHelper
from databricks.feature_store._catalog_client_helper import CatalogClientHelper

from pyspark.sql import DataFrame
from pyspark.sql.utils import AnalysisException
from pyspark.sql.types import StructType
from pyspark.sql.streaming import StreamingQuery

_logger = logging.getLogger(__name__)


class ComputeClient:
    """
    The compute client manages metadata about feature tables, eg:

    - Creating/registering feature tables
    - Reading feature table metadata
    - Dropping feature tables from the catalog
    - Managing attributes of feature tables such as tags
    """

    _WRITE_MODES = [OVERWRITE, MERGE]

    def __init__(
        self,
        catalog_client: CatalogClient,
        catalog_client_helper: CatalogClientHelper,
        hive_client: HiveClient,
    ):
        self._catalog_client = catalog_client
        self._catalog_client_helper = catalog_client_helper
        self._hive_client = hive_client
        self._hive_client_helper = HiveClientHelper(self._hive_client)

    def create_table(
        self,
        name: str,
        primary_keys: Union[str, List[str]],
        df: Optional[DataFrame] = None,
        *,
        timestamp_keys: Union[str, List[str], None] = None,
        partition_columns: Union[str, List[str], None] = None,
        schema: Optional[StructType] = None,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> FeatureTable:

        features_df = kwargs.pop("features_df", None)
        if features_df is not None and df is not None:
            raise ValueError("Either features_df or df can be provided, but not both.")
        if features_df is not None:
            _logger.warning(
                'The "features_df" parameter is deprecated. Use "df" instead.'
            )
            df = features_df
        path = kwargs.pop("path", None)
        databricks.feature_store.utils.validation_utils.check_kwargs_empty(
            kwargs, "create_table"
        )

        return self._create_table(
            name,
            primary_keys,
            df,
            timestamp_keys=timestamp_keys,
            partition_columns=partition_columns,
            schema=schema,
            description=description,
            path=path,
            tags=tags,
            req_context=RequestContext(request_context.CREATE_TABLE),
        )

    def _create_table(
        self,
        name: str,
        primary_keys: Union[str, List[str]],
        df: DataFrame = None,
        *,
        timestamp_keys: Union[str, List[str]] = None,
        partition_columns: Union[str, List[str]] = None,
        schema: StructType = None,
        description: str = None,
        path: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        req_context: Optional[RequestContext] = None,
    ) -> FeatureTable:

        if schema is None and df is None:
            raise ValueError("Either schema or df must be provided")

        if schema and df and not ComputeClient._schema_eq(schema, df.schema):
            raise ValueError(
                "Provide either 'schema' or 'df' arguments. If both arguments "
                "are provided, their schemas must match."
            )

        if df is not None:
            databricks.feature_store.utils.validation_utils.check_dataframe_type(df)
        databricks.feature_store.utils.validation_utils.check_qualified_name(name)
        self._hive_client_helper.check_database_exists(name)

        table_schema = schema or df.schema
        ComputeClient._check_schema_top_level_types_supported(table_schema)

        partition_cols_as_list = utils.as_list(partition_columns, default=[])
        if partition_columns:
            ComputeClient._check_schema_has_columns(
                table_schema, partition_cols_as_list, "partition columns"
            )

        primary_keys_as_list = utils.as_list(primary_keys)
        ComputeClient._check_schema_has_columns(
            table_schema, primary_keys_as_list, "primary keys"
        )

        timestamp_keys_as_list = utils.as_list(timestamp_keys, default=[])
        if timestamp_keys:
            ComputeClient._check_schema_has_columns(
                table_schema, timestamp_keys_as_list, "timestamp keys"
            )

        # 1. Handle cases where the table exists in either Hive or the Catalog
        delta_table_exists = self._hive_client.table_exists(name)
        catalog_table_exists = self._catalog_client.feature_table_exists(
            name, req_context
        )

        if delta_table_exists and not catalog_table_exists:
            raise ValueError(f"Data table {name} already exists. Use a different name.")

        if catalog_table_exists and not delta_table_exists:
            raise ValueError(
                f"Feature table {name} already exists, but data table not accessible in Spark. "
                f"Consider deleting the feature table to resolve this error."
            )

        if catalog_table_exists and delta_table_exists:
            return self._check_catalog_matches_delta_metadata(
                name,
                table_schema,
                primary_keys_as_list,
                partition_cols_as_list,
                timestamp_keys_as_list,
                req_context,
            )

        # At this point, neither the Delta table nor the Catalog table exist.

        # 2. Create empty Delta table. If this fails for some reason, the Feature Table will not be
        # added to the Feature Catalog.
        self._hive_client.create_table(name, table_schema, partition_cols_as_list, path)

        # 3. Add feature table and features to the Feature Catalog.
        # Features (other than primary keys and partition keys) are added in a separate call.
        delta_schema = {
            feature.name: feature.dataType
            for feature in self._hive_client.get_feature_table_schema(name)
        }

        # TODO(ML-17484): Move the following checks prior to Hive table creation

        partition_key_specs = []
        for k in partition_cols_as_list:
            spark_data_type = delta_schema[k]
            partition_key_specs.append(KeySpec(k, spark_data_type.typeName()))

        primary_key_specs = []
        for k in primary_keys_as_list:
            spark_data_type = delta_schema[k]
            primary_key_specs.append(KeySpec(k, spark_data_type.typeName()))

        timestamp_key_specs = []
        for k in timestamp_keys_as_list:
            spark_data_type = delta_schema[k]
            timestamp_key_specs.append(KeySpec(k, spark_data_type.typeName()))

        feature_key_specs = self._get_feature_key_specs(
            delta_schema,
            primary_keys_as_list,
            timestamp_keys_as_list,
            partition_cols_as_list,
        )

        try:
            self._create_feature_table_with_features_and_tags(
                name=name,
                partition_key_specs=partition_key_specs,
                primary_key_specs=primary_key_specs,
                timestamp_key_specs=timestamp_key_specs,
                description=description,
                is_imported=False,
                feature_key_specs=feature_key_specs,
                tags=tags,
                req_context=req_context,
            )
        except Exception as e:
            # Delete empty Delta table.  The feature table will have already been cleaned up from the catalog.
            self._hive_client.delete_empty_table(name)
            raise e

        # 4. Write to Delta table
        if df is not None:
            try:
                # Use mode OVERWRITE since this a new feature table.
                self.write_table(
                    name,
                    df,
                    mode=OVERWRITE,
                    producer_action=ProducerAction.CREATE,
                    req_context=req_context,
                )
            except Exception as e:
                # Delete the entire delta table if fatal exception occurs.
                # This may happen after partial data was written and an unknown exception is thrown.
                # It is OK to delete the feature table here because we are certain this is the
                # feature table we just created. We should NOT delete any existing feature table
                # created by user
                self._hive_client.delete_table(name)
                self._catalog_client.delete_feature_table(name, req_context)
                raise e

        _logger.info(f"Created feature table '{name}'.")
        return self.get_table(name, req_context)

    def register_table(
        self,
        *,
        delta_table: str,
        primary_keys: Union[str, List[str]],
        timestamp_keys: Union[str, List[str], None] = None,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
    ) -> FeatureTable:

        # Validate if the provided Delta table exists.
        req_context = RequestContext(request_context.REGISTER_TABLE)
        if not is_qualified_table_name(delta_table):
            raise ValueError(
                "The provided Delta table name must have the form <database_name>.<table_name> "
                "and only contain alphabet characters, numbers and _."
            )
        self._hive_client_helper.check_database_exists(delta_table)
        if not self._hive_client.table_exists(delta_table):
            raise ValueError(
                f"The provided Delta table '{delta_table}' could not be found."
            )
        df = self._hive_client.read_table(delta_table)

        # Validate if the provided Delta table is feature store compliant
        # 1. Check if the Delta table contains valid types and the specified primary key and
        #    timestamp key columns.
        table_schema = df.schema
        ComputeClient._check_schema_top_level_types_supported(table_schema)

        primary_keys_as_list = utils.as_list(primary_keys)
        ComputeClient._check_schema_has_columns(
            table_schema, primary_keys_as_list, "primary keys"
        )

        timestamp_keys_as_list = utils.as_list(timestamp_keys, default=[])
        if timestamp_keys:
            ComputeClient._check_schema_has_columns(
                table_schema, timestamp_keys_as_list, "timestamp keys"
            )

        # 2. Check that the Delta table does not contain generated columns.
        #    More details: go/import_delta_table
        if self._hive_client.has_generated_columns(delta_table):
            raise ValueError(
                f"Provided Delta table must not contain generated column(s)."
            )
        # 3. Handle the case where the Delta table is already a feature table.
        if self._catalog_client.feature_table_exists(delta_table, req_context):
            return self._check_catalog_matches_delta_metadata(
                delta_table,
                table_schema,
                primary_keys_as_list,
                [],
                timestamp_keys_as_list,
                req_context,
            )

        # 4. Check no two rows have the same primary keys.
        if self._hive_client.df_violates_pk_constraint(
            df, primary_keys_as_list + timestamp_keys_as_list
        ):
            raise ValueError(
                f"Non-unique rows detected in input dataframe for key combination"
                f"{ primary_keys_as_list + timestamp_keys_as_list}."
            )

        # Register the table with feature store
        delta_schema = {feature.name: feature.dataType for feature in table_schema}
        primary_key_specs = []
        for k in primary_keys_as_list:
            spark_data_type = delta_schema[k]
            primary_key_specs.append(KeySpec(k, spark_data_type.typeName()))

        timestamp_key_specs = []
        for k in timestamp_keys_as_list:
            spark_data_type = delta_schema[k]
            timestamp_key_specs.append(KeySpec(k, spark_data_type.typeName()))

        feature_key_specs = self._get_feature_key_specs(
            delta_schema,
            primary_keys_as_list,
            timestamp_keys_as_list,
            [],
        )

        self._create_feature_table_with_features_and_tags(
            name=delta_table,
            partition_key_specs=[],
            primary_key_specs=primary_key_specs,
            timestamp_key_specs=timestamp_key_specs,
            description=description,
            is_imported=True,
            feature_key_specs=feature_key_specs,
            tags=tags,
            req_context=req_context,
        )

        self._catalog_client_helper.add_producers(
            delta_table, ProducerAction.REGISTER, req_context
        )
        return self.get_table(delta_table, req_context)

    def write_table(
        self,
        name: str,
        df: DataFrame,
        mode: str,
        req_context: RequestContext,
        checkpoint_location: Union[str, None] = None,
        trigger: Dict[str, Any] = _DEFAULT_WRITE_STREAM_TRIGGER,
        producer_action: ProducerAction = None,
    ) -> Union[StreamingQuery, None]:
        """
        Write to a feature table.

        If the input :class:`DataFrame <pyspark.sql.DataFrame>` is streaming, will create a write stream.

        :param name: A feature table name of the form ``<database_name>.<table_name>``,
          for example ``dev.user_features``. Raises an exception if this feature table does not
          exist.
        :param df: Spark :class:`DataFrame <pyspark.sql.DataFrame>` with feature data. Raises an exception if the schema does not
          match that of the feature table.
        :param mode: Two supported write modes:

          * ``"overwrite"`` updates the whole table.

          * ``"merge"`` will upsert the rows in ``df`` into the feature table. If ``df`` contains
            columns not present in the feature table, these columns will be added as new features.

        :param checkpoint_location: Sets the Structured Streaming ``checkpointLocation`` option.
          By setting a ``checkpoint_location``, Spark Structured Streaming will store
          progress information and intermediate state, enabling recovery after failures.
          This parameter is only supported when the argument ``df`` is a streaming :class:`DataFrame <pyspark.sql.DataFrame>`.
        :param trigger: If ``df.isStreaming``, ``trigger`` defines the timing of stream data
          processing, the dictionary will be unpacked and passed to :meth:`DataStreamWriter.trigger <pyspark.sql.streaming.DataStreamWriter.trigger>`
          as arguments. For example, ``trigger={'once': True}`` will result in a call to
          ``DataStreamWriter.trigger(once=True)``.
        :return: If ``df.isStreaming``, returns a PySpark :class:`StreaminQuery <pyspark.sql.streaming.StreamingQuery>`, :obj:`None` otherwise.
        """
        databricks.feature_store.utils.validation_utils.check_dataframe_type(df)
        mode_string = mode.strip().lower()
        if mode_string not in self._WRITE_MODES:
            supported_modes_list = ", ".join([f"'{m}'" for m in self._WRITE_MODES])
            raise ValueError(
                f"Unsupported mode '{mode}'. Use one of {supported_modes_list}"
            )

        checkpoint_location = databricks.feature_store.utils.validation_utils.standardize_checkpoint_location(
            checkpoint_location
        )
        if checkpoint_location is not None and not df.isStreaming:
            _logger.warning("Ignoring checkpoint_location, since df is not streaming.")
            checkpoint_location = None

        ComputeClient._check_schema_top_level_types_supported(df.schema)
        self._hive_client_helper.check_feature_table_exists_in_hive(name)
        feature_table = self._catalog_client.get_feature_table(name, req_context)

        # We know from the successful `get_feature_table call` above that the user has
        # at least read permission. Otherwise backend will throw RESOURCE_DOES_NOT_EXIST exception.
        # Since this is a write operation, we want to check whether the user has write permission
        # on the feature table prior to other operations.
        if not self._catalog_client.can_write_to_catalog(name, req_context):
            raise PermissionError(
                f"You do not have permission to write to feature table {name}."
            )

        # Validate: Internal state is consistent. Existing Delta schema should match Catalog schema.
        features = self._catalog_client.get_features(name, req_context)
        existing_schema = self._hive_client.get_feature_table_schema(name)
        if not schema_utils.catalog_matches_delta_schema(features, existing_schema):
            # If the existing Delta table does not match the Feature Catalog, the state is invalid
            # and we cannot write to the feature table. Error out.
            schema_utils.log_catalog_schema_not_match_delta_schema(
                features, existing_schema, level=_ERROR
            )

        # Validate: Provided DataFrame has key and partition columns
        features = self._catalog_client.get_features(name, req_context)
        if not schema_utils.catalog_matches_delta_schema(
            features, df.schema, column_filter=feature_table.primary_keys
        ):
            raise ValueError(
                f"The provided DataFrame must contain all specified primary key columns and have "
                f"the same type. Could not find key(s) '{feature_table.primary_keys}' with "
                f"correct types in schema {df.schema}."
            )
        if not schema_utils.catalog_matches_delta_schema(
            features, df.schema, column_filter=feature_table.partition_columns
        ):
            raise ValueError(
                f"The provided DataFrame must contain all specified partition columns and have "
                f"the same type. Could not find partition column(s) "
                f"'{feature_table.partition_columns}' with correct types in schema {df.schema}."
            )
        if not schema_utils.catalog_matches_delta_schema(
            features, df.schema, column_filter=feature_table.timestamp_keys
        ):
            raise ValueError(
                f"The provided DataFrame must contain the specified timestamp key column "
                f"and have the same type. Could not find key '{feature_table.timestamp_keys[0]}' "
                f"with correct types in schema {df.schema}."
            )

        # Invariant: We know from a validation check above that the Delta table schema matches the
        # Catalog schema.

        # Check for schema differences between the Catalog feature table and df's schema.
        if not schema_utils.catalog_matches_delta_schema(features, df.schema):
            # If this is a feature table with point-in-time lookup timestamp keys.
            # Validate: all existing table columns are present in the df.
            if feature_table.timestamp_keys:
                feature_names = [feature.name for feature in features]
                df_column_names = [c.name for c in df.schema]
                missing_column_names = list(set(feature_names) - set(df_column_names))
                if missing_column_names:
                    raise ValueError(
                        f"Feature table has a timestamp column. When calling write_table "
                        f"the provided DataFrame must contain all the feature columns. "
                        f"Could not find column(s) '{missing_column_names}'."
                    )
            # Attempt to update both the Delta table and Catalog schemas.
            # New columns will be added, column type mismatch will raise an error.
            ComputeClient._check_unique_case_insensitive_schema(features, df.schema)
            # First update the Delta schema. Spark will handle any type changes, and throw on
            # incompatible types.
            self._update_delta_features(name, df.schema)
            # Now update the Catalog using *the types in the Delta table*. We do not use the types
            # in `df` here so we can defer schema merging logic to Spark.
            delta_schema = self._hive_client.get_feature_table_schema(name)
            self._update_catalog_features_with_delta_schema(
                name, feature_table, features, delta_schema, req_context
            )

        # Exclude self-referential data sources. Feature table should exist.
        feature_table_data_source = self._hive_client.get_delta_table_path(name)
        # set(None) can produce exception, set default to empty list
        excluded_paths = set(utils.as_list(feature_table_data_source, default=[]))

        # Write data to Delta table
        with SparkSourceListener() as spark_source_listener:
            return_value = self._hive_client.write_table(
                name,
                feature_table.primary_keys,
                feature_table.timestamp_keys,
                df,
                mode_string,
                checkpoint_location,
                trigger,
            )
            subscribed_data_sources = spark_source_listener.get_data_sources()

        tables = set()
        paths = set()
        for fmt, sources in subscribed_data_sources.items():
            # filter out source that are an exact match to the excluded paths.
            # ToDo(mparkhe): Currently Spark listener will not return subdirs as data sources,
            #                but in future investigate a clean mechanism to deduplicate,
            #                and eliminate redundant subdirs reported as (Delta) data sources.
            #                eg: ["dbfs:/X.db/Y", "dbfs:/X.db/Y/_delta_log/checkpoint..."]
            valid_sources = list(
                filter(
                    lambda source: source not in excluded_paths,
                    sources,
                )
            )
            if len(valid_sources) > 0:
                # We rely on the spark listener to determine whether a data source is a delta table
                # for now. However, spark listener categorize delta table by looking up the
                # leaf node in spark query plan whereas we categorize delta table by whether or not
                # it would show up in the `Data` tab. Inconsistency could happen if user reads a
                # delta directory as delta table through `spark.read.format("delta")`,
                # we should store such data source as a path rather than a delta table.
                if fmt == _SOURCE_FORMAT_DELTA:
                    for path in valid_sources:
                        # Convert table-paths to "db_name.table_name".
                        # Note: If a table-path does not match the top level DBFS path
                        #       it is preserved as is.
                        converted_table = self._hive_client.convert_to_table_format(
                            path
                        )
                        if converted_table == path:
                            # Failed to convert table-path to "db_name.table_name",
                            # record data source as a path
                            paths.add(path)
                        else:
                            tables.add(converted_table)
                            # Exclude DBFS paths for all the table data sources
                            excluded_paths.add(path)
                else:
                    paths.update(valid_sources)

        # filter out paths match or are subdirectory (or files) under excluded paths
        # Example: if excluded_paths = ["dbfs:/path/to/database.db/table]
        #          also exclude sub-paths like "dbfs:/path/to/database.db/table/subpath"
        #          but do not exclude "dbfs:/path/to/database.db/tablesubdir"
        # ToDo(mparkhe): In future investigate a clean mechanism to eliminate subdirs
        #                of path sources, if returned by Spark listener.
        #                eg: ["dbfs:/X/Y", "dbfs:/X/Y/subdir"] => ["dbfs:/X/Y"]
        valid_paths = list(
            filter(
                lambda source: all(
                    [
                        source != excluded_path
                        and not source.startswith(utils.as_directory(excluded_path))
                        for excluded_path in excluded_paths
                    ]
                ),
                paths,
            )
        )

        # record data sources to feature catalog
        if len(tables) > 0 or len(valid_paths) > 0:
            self._catalog_client_helper.add_data_sources(
                name=name,
                tables=tables,
                paths=valid_paths,
                custom_sources=set(),  # No custom_sources in auto tracked data sources
                req_context=req_context,
            )
        # record producer to feature catalog
        self._catalog_client_helper.add_producers(name, producer_action, req_context)

        return return_value

    def get_table(self, name: str, req_context: RequestContext) -> FeatureTable:
        self._hive_client_helper.check_feature_table_exists_in_hive(name)
        feature_table = self._catalog_client.get_feature_table(name, req_context)
        features = self._catalog_client.get_features(name, req_context)
        df = self._hive_client.read_table(name)
        if not schema_utils.catalog_matches_delta_schema(features, df.schema):
            schema_utils.log_catalog_schema_not_match_delta_schema(
                features, df.schema, level=_WARN
            )
        tag_entities = self._catalog_client.get_feature_table_tags(
            feature_table.table_id, req_context
        )
        feature_table._tags = {
            tag_entity.key: tag_entity.value for tag_entity in tag_entities
        }
        return feature_table

    def drop_table(self, name: str) -> None:
        req_context = RequestContext(request_context.DROP_TABLE)
        databricks.feature_store.utils.validation_utils.check_qualified_name(name)

        delta_table_exist = self._hive_client.table_exists(name)
        feature_table_exists = self._catalog_client.feature_table_exists(
            name, req_context
        )

        # Handle cases where catalog data does not exist.
        if not feature_table_exists and delta_table_exist:
            raise ValueError(
                f"Delta table '{name}' is not a feature table. Use spark API to drop the delta table. "
                f"For more information on Spark API, "
                f"see https://docs.databricks.com/sql/language-manual/sql-ref-syntax-ddl-drop-table.html."
            )
        if not feature_table_exists and not delta_table_exist:
            raise ValueError(f"Feature table '{name}' does not exist.")

        feature_table = self._catalog_client.get_feature_table(name, req_context)

        # Delete the feature table and underlying delta table.
        # First perform a dry-run deletion of catalog data as the backend validates the API call.
        try:
            self._catalog_client.delete_feature_table(name, req_context, dry_run=True)
        except Exception as e:
            _logger.error(f"Unable to delete the feature table due to {e}.")
            raise e
        self._hive_client.delete_table(name)
        try:
            self._catalog_client.delete_feature_table(name, req_context)
        except Exception as e:
            _logger.error(
                f"Failed to delete the feature table from Feature Catalog due to {e}."
                f" To fix this, re-run the 'drop_table' method."
            )
            raise e
        _logger.warning(
            "Deleting a feature table can lead to unexpected failures in upstream "
            "producers and downstream consumers (models, endpoints, and scheduled jobs)."
        )
        if feature_table.online_stores:
            ComputeClient._log_online_store_info(feature_table.online_stores)

    def read_table(self, name: str, **kwargs) -> DataFrame:
        as_of_delta_timestamp = kwargs.pop("as_of_delta_timestamp", None)
        databricks.feature_store.utils.validation_utils.check_kwargs_empty(
            kwargs, "read_table"
        )

        req_context = RequestContext(request_context.READ_TABLE)
        self._hive_client_helper.check_feature_table_exists_in_hive(name)
        df = self._hive_client.read_table(name, as_of_delta_timestamp)
        features = self._catalog_client.get_features(name, req_context)
        if not schema_utils.catalog_matches_delta_schema(features, df.schema):
            schema_utils.log_catalog_schema_not_match_delta_schema(
                features, df.schema, level=_WARN
            )
        # Add consumer of each feature as final step
        consumer_feature_table_map = {name: [feature.name for feature in features]}
        self._catalog_client_helper.add_consumers(
            consumer_feature_table_map, req_context
        )
        return df

    def set_feature_table_tag(self, *, table_name: str, key: str, value: str) -> None:
        utils.validate_params_non_empty(locals(), ["table_name", "key", "value"])
        req_context = RequestContext(request_context.SET_FEATURE_TABLE_TAG)
        ft = self.get_table(table_name, req_context)
        self._catalog_client.set_feature_table_tags(
            ft.table_id, {key: value}, req_context
        )

    def delete_feature_table_tag(self, *, table_name: str, key: str) -> None:
        utils.validate_params_non_empty(locals(), ["table_name", "key"])
        req_context = RequestContext(request_context.DELETE_FEATURE_TABLE_TAG)
        ft = self.get_table(table_name, req_context)
        if key not in ft.tags:
            _logger.warning(
                f'The tag "{key}" for feature table "{table_name}" was not found, so the delete operation has been skipped.'
            )
        else:
            self._catalog_client.delete_feature_table_tags(
                ft.table_id, [key], req_context
            )

    def _check_catalog_matches_delta_metadata(
        self,
        name,
        table_schema,
        primary_keys_as_list,
        partition_cols_as_list,
        timestamp_keys_as_list,
        req_context,
    ) -> FeatureTable:
        """
        Checks if existing feature table catalog metadata with {name} matches the data table
        metadata including the table_schema, primary keys, timestamp keys and partition columns.
        Return the existing feature table if there is a match, otherwise raise an error.
        """
        ft = self.get_table(name, req_context)
        existing_features = self._catalog_client.get_features(name, req_context)
        schemas_match = schema_utils.catalog_matches_delta_schema(
            existing_features, table_schema
        )
        primary_keys_match = primary_keys_as_list == ft.primary_keys
        partition_keys_match = partition_cols_as_list == ft.partition_columns
        timestamp_keys_match = timestamp_keys_as_list == ft.timestamp_keys
        if (
            schemas_match
            and primary_keys_match
            and partition_keys_match
            and timestamp_keys_match
        ):
            _logger.warning(
                f'The feature table "{name}" already exists. Use "FeatureStoreClient.write_table"'
                f" API to write to the feature table."
            )
            return ft
        else:
            error_msg = (
                f"The feature table '{name}' already exists with a different schema.:\n"
            )
            if not schemas_match:
                error_msg += (
                    f"Existing schema: {existing_features}\n"
                    f"New schema:{table_schema}\n\n"
                )
            if not primary_keys_match:
                error_msg += (
                    f"Existing primary keys: {ft.primary_keys}\n"
                    f"New primary keys: {primary_keys_as_list}\n\n"
                )
            if not partition_keys_match:
                error_msg += (
                    f"Existing partition keys: {ft.partition_columns}\n"
                    f"New partition keys: {partition_cols_as_list}\n\n"
                )
            if not timestamp_keys_match:
                error_msg += (
                    f"Existing timestamp keys: {ft.timestamp_keys}\n"
                    f"New timestamp keys: {timestamp_keys_as_list}\n\n"
                )

            raise ValueError(error_msg)

    @staticmethod
    def _schema_eq(schema1, schema2):
        return set(schema1.fields) == set(schema2.fields)

    @staticmethod
    def _check_schema_top_level_types_supported(schema: StructType) -> None:
        """
        Checks whether the provided schema is supported by Feature Store, only considering the
        top-level type for nested data types.
        """
        unsupported_name_type = [
            (field.name, field.dataType)
            for field in schema.fields
            if not DataType.top_level_type_supported(field.dataType)
        ]
        if unsupported_name_type:
            plural = len(unsupported_name_type) > 1
            missing_cols_str = ", ".join(
                [
                    f"\n\t- {feat_name} (type: {feat_type})"
                    for (feat_name, feat_type) in unsupported_name_type
                ]
            )
            raise ValueError(
                f"Unsupported data type for column{'s' if plural else ''}: {missing_cols_str}"
            )

    @staticmethod
    def _check_schema_has_columns(schema, columns, col_type):
        schema_cols = [field.name for field in schema.fields]
        for col in columns:
            if col not in schema_cols:
                raise ValueError(
                    f"The provided DataFrame or schema must contain all specified {col_type}. "
                    f"Schema {schema} is missing column '{col}'"
                )

    @staticmethod
    def _get_feature_key_specs(
        delta_schema: StructType,
        primary_keys_as_list: List[str],
        timestamp_keys_as_list: List[str],
        partition_cols_as_list: List[str],
    ) -> List[KeySpec]:
        """
        Returns the KeySpec for only features in the delta_schema. KeySpecs are not created for
        primary keys, partition keys, and timestamp keys.
        """
        feature_key_specs = []
        for k in delta_schema:
            if (
                k not in partition_cols_as_list
                and k not in primary_keys_as_list
                and k not in timestamp_keys_as_list
            ):
                spark_data_type = delta_schema[k]
                # If the feature is a complex Spark DataType, convert the Spark DataType to its
                # JSON representation to be updated in the Feature Catalog.
                data_type_details = (
                    spark_data_type.json()
                    if DataType.from_spark_type(spark_data_type)
                    in DATA_TYPES_REQUIRES_DETAILS
                    else None
                )
                feature_key_specs.append(
                    KeySpec(k, spark_data_type.typeName(), data_type_details)
                )
        return feature_key_specs

    def _create_feature_table_with_features_and_tags(
        self,
        *,
        name: str,
        partition_key_specs: List[KeySpec],
        primary_key_specs: List[KeySpec],
        timestamp_key_specs: List[KeySpec],
        feature_key_specs: List[KeySpec],
        is_imported: bool,
        tags: Optional[Dict[str, str]] = None,
        description: str = None,
        req_context: Optional[RequestContext] = None,
    ) -> FeatureTable:
        """
        Create the feature_table, features and tags.

        If any step fails, the exception handler cleans up the feature table from the feature catalog
        and propagates the exception to the caller for further handling.
        """
        feature_table = None
        try:
            feature_table = self._catalog_client.create_feature_table(
                name,
                partition_key_spec=partition_key_specs,
                primary_key_spec=primary_key_specs,
                timestamp_key_spec=timestamp_key_specs,
                description=description,
                is_imported=is_imported,
                req_context=req_context,
            )
            if len(feature_key_specs) > 0:
                self._catalog_client.create_features(
                    name, feature_key_specs, req_context
                )
            if tags:
                self._catalog_client.set_feature_table_tags(
                    feature_table.table_id, tags, req_context
                )
            return feature_table
        except Exception as e:
            # Delete the newly created feature table in the catalog
            if feature_table:
                self._catalog_client.delete_feature_table(name, req_context)
            raise e

    # TODO(ML-21475): Move this logic to entities class
    @staticmethod
    def _log_online_store_info(online_stores):
        message = "You must delete the following published online stores with your cloud provider: \n"
        for online_store in online_stores:
            message += f"\t - '{online_store.name}' ({utils.get_canonical_online_store_name(online_store)}) \n"
            if online_store.store_type == StoreType.DYNAMODB:
                message += f"\t\t - Region: {online_store.dynamodb_metadata.region}, Table_arn: {online_store.dynamodb_metadata.table_arn} \n"
            elif online_store.store_type == StoreType.MYSQL:
                message += f"\t\t - Host: {online_store.mysql_metadata.host}, Port: {online_store.mysql_metadata.port} \n"
            elif online_store.store_type == StoreType.SQL_SERVER:
                message += f"\t\t - Host: {online_store.sql_server_metadata.host}, Port: {online_store.sql_server_metadata.port} \n"
            else:
                message += "Unknown online store."
        _logger.warning(message)

    @staticmethod
    def _check_unique_case_insensitive_schema(
        catalog_features: List[Feature], df_schema: DataFrame
    ) -> None:
        """
        Verify schema is unique and case sensitive.

        Confirm that column names in Feature Catalog and user's input
        DataFrame do not have duplicate
        case insensitive columns when writing data to the feature table.

        Prevents the following cases:
        1. User input DataFrame's schema is '{'feat1': 'FLOAT', 'FEAT1': 'FLOAT'}'
        2. User input DataFrame's schema is '{'FEAT1': 'FLOAT'}', and Feature Catalog's schema is
        '{'feat1': 'FLOAT'}'
        """
        df_cols = {}
        for df_column in df_schema:
            if df_column.name.lower() in df_cols:
                raise ValueError(
                    f"The provided DataFrame cannot contain duplicate column names. Column names are case insensitive. "
                    f"The DataFrame contains duplicate columns: {df_cols[df_column.name.lower()]}, {df_column.name}"
                )
            df_cols[df_column.name.lower()] = df_column.name

        for feature in catalog_features:
            if (
                feature.name.lower() in df_cols
                and feature.name != df_cols[feature.name.lower()]
            ):
                raise ValueError(
                    f"Feature names cannot differ by only case. The provided DataFrame has column "
                    f"{df_cols[feature.name.lower()]}, which duplicates the Feature Catalog column {feature.name}. "
                    f"Please rename the column"
                )

    def _update_delta_features(self, name, schema):
        """
        Update the Delta table with name `name`.

        This update happens by merging in `schema`. Will throw if the schema
        is incompatible with the existing Delta table schema.

        .. note::

           Validate: Delta table schemas are compatible. Because HiveClient.write_table enables
           the "mergeSchema" option, differences in schema will be reconciled by Spark. We will
           later write this schema to the Feature Catalog. In this way, we defer the schema
           merging logic to Spark.
        """
        try:
            self._hive_client.attempt_to_update_delta_table_schema(name, schema)
        except AnalysisException as e:
            raise ValueError(
                "FeatureStoreClient uses Delta APIs. The schema of the new DataFrame is "
                f"incompatible with existing Delta table. Saw AnalysisException: {str(e)}"
            )

    def _update_catalog_features_with_delta_schema(
        self, name, ft, features, delta_schema, req_context: RequestContext
    ):
        """
        Update the catalog to include all columns of the provided Delta table schema.

        :param name: Feature table name
        :param ft: FeatureTable
        :param features: [Features]
        :param delta_schema: Schema of the data table.
        :param req_context: The RequestContext
        """
        catalog_features_to_fs_types = {
            f.name: DataType.from_string(f.data_type) for f in features
        }
        delta_features_to_fs_types = {
            feature.name: DataType.from_spark_type(feature.dataType)
            for feature in delta_schema
        }
        complex_catalog_features_to_spark_types = (
            schema_utils.get_complex_catalog_schema(
                features, catalog_features_to_fs_types
            )
        )
        complex_delta_features_to_spark_types = schema_utils.get_complex_delta_schema(
            delta_schema, delta_features_to_fs_types
        )

        feaures_and_data_types_to_add = []
        features_and_data_types_to_update = []
        for feat, fs_data_type in delta_features_to_fs_types.items():
            simple_types_mismatch = (feat in catalog_features_to_fs_types) and (
                fs_data_type != catalog_features_to_fs_types[feat]
            )
            complex_types_mismatch = (
                feat in complex_catalog_features_to_spark_types
            ) and (
                complex_catalog_features_to_spark_types[feat]
                != complex_delta_features_to_spark_types[feat]
            )
            if simple_types_mismatch or complex_types_mismatch:
                # If the feature is a complex Spark DataType, convert the Spark DataType to its
                # JSON representation to be updated in the Feature Catalog.
                data_type_details = complex_delta_features_to_spark_types.get(feat)
                if data_type_details:
                    data_type_details = data_type_details.json()
                features_and_data_types_to_update.append(
                    (feat, fs_data_type, data_type_details)
                )
            if feat not in ft.primary_keys and feat not in catalog_features_to_fs_types:
                # If the feature is a complex Spark DataType, convert the Spark DataType to its
                # JSON representation to be updated in the Feature Catalog.
                data_type_details = complex_delta_features_to_spark_types.get(feat)
                if data_type_details:
                    data_type_details = data_type_details.json()
                feaures_and_data_types_to_add.append(
                    (feat, fs_data_type, data_type_details)
                )
        if feaures_and_data_types_to_add:
            key_specs = [
                KeySpec(
                    feat,
                    DataType.to_string(data_type),
                    data_type_details,
                )
                for (
                    feat,
                    data_type,
                    data_type_details,
                ) in feaures_and_data_types_to_add
            ]
            self._catalog_client.create_features(name, key_specs, req_context)
        # There is no need to update types of existing columns because mergeSchema does not support
        # column type changes.



import logging
from typing import List, Union, Dict, Any, Optional, Set
from types import ModuleType

from databricks.feature_store.api.proto.feature_catalog_pb2 import ProducerAction
from databricks.feature_store._compute_client._compute_client import ComputeClient
from databricks.feature_store._hive_client_helper import HiveClientHelper
from databricks.feature_store._catalog_client_helper import CatalogClientHelper
from databricks.feature_store.constants import (
    OVERWRITE,
    MERGE,
    CUSTOM,
    TABLE,
    PATH,
    _DEFAULT_WRITE_STREAM_TRIGGER,
    _DEFAULT_PUBLISH_STREAM_TRIGGER,
    _SOURCE_FORMAT_DELTA,
    _WARN,
    _ERROR,
)
from databricks.feature_store.entities.feature_lookup import FeatureLookup
from databricks.feature_store.entities.feature import Feature
from databricks.feature_store.entities.feature_column_info import FeatureColumnInfo
from databricks.feature_store.entities.feature_table import FeatureTable
from databricks.feature_store.entities.source_data_column_info import (
    SourceDataColumnInfo,
)
from databricks.feature_store.entities.online_store_metadata import (
    OnlineStoreMetadata,
    augment_online_store,
)
from databricks.feature_store.hive_client import HiveClient
from databricks.feature_store.catalog_client import CatalogClient
from databricks.feature_store.databricks_client import DatabricksClient
from databricks.feature_store.online_store_spec import (
    OnlineStoreSpec,
)
from databricks.feature_store.online_store_spec.online_store_properties import (
    HOSTNAME,
    PORT,
    DATABASE_NAME,
    TABLE_NAME,
)
from databricks.feature_store.training_set import TrainingSet
from databricks.feature_store.utils import utils
from databricks.feature_store.utils.publish_utils import (
    update_online_store_spec_sticky_ttl,
)
from databricks.feature_store.utils.rest_utils import get_error_code
from databricks.feature_store.utils.feature_lookup_utils import join_feature_data
from databricks.feature_store.utils import request_context
from databricks.feature_store.utils.request_context import (
    RequestContext,
)
from databricks.feature_store.utils import schema_utils
from databricks.feature_store.utils import validation_utils
from databricks.feature_store._training_scoring_client._training_scoring_client import (
    TrainingScoringClient,
)
from databricks.feature_store._publish_client._publish_client import (
    PublishClient,
)

from pyspark.sql import DataFrame
from pyspark.sql.types import StructType
from pyspark.sql.streaming import StreamingQuery

from mlflow.utils import databricks_utils
from mlflow.utils.annotations import deprecated, experimental
import mlflow

_logger = logging.getLogger(__name__)


class FeatureStoreClient:
    """Client for interacting with the Databricks Feature Store."""

    _WRITE_MODES = [OVERWRITE, MERGE]
    _PUBLISH_MODES = [OVERWRITE, MERGE]
    _DATA_SOURCE_TYPES = [CUSTOM, PATH, TABLE]
    _DEFAULT_WRITE_STREAM_TRIGGER = _DEFAULT_WRITE_STREAM_TRIGGER
    _DEFAULT_PUBLISH_STREAM_TRIGGER = _DEFAULT_PUBLISH_STREAM_TRIGGER
    _SOURCE_FORMAT_DELTA = _SOURCE_FORMAT_DELTA
    _WARN = _WARN
    _ERROR = _ERROR

    def __init__(
        self,
        feature_store_uri: Optional[str] = None,
        model_registry_uri: Optional[str] = None,
    ):
        """
        Initialize a client to interact with the feature store.

        Creates a client to interact with the feature store. Takes in an optional parameter to identify the remote
        workspace for multi-workspace Feature Store.

        :param feature_store_uri: An URI of the form ``databricks://<scope>.<prefix>`` that identifies the credentials
          of the intended Feature Store workspace. Throws an error if specified but credentials were not found.
        :param model_registry_uri: Address of local or remote model registry server. If not provided,
          defaults to the local server.
        """

        if not utils.is_in_databricks_env():
            _logger.warning(
                f"The Databricks Feature Store client is intended to be run on Databricks. "
                f"Local and external development is not supported."
            )

        self._catalog_client = CatalogClient(
            databricks_utils.get_databricks_host_creds, feature_store_uri
        )
        # The Databricks client must be local from the context of the notebook
        self._databricks_client = DatabricksClient(
            databricks_utils.get_databricks_host_creds
        )
        self._hive_client = HiveClient()
        self._hive_client_helper = HiveClientHelper(self._hive_client)
        self._catalog_client_helper = CatalogClientHelper(
            self._catalog_client, self._databricks_client
        )
        self._compute_client = ComputeClient(
            catalog_client=self._catalog_client,
            catalog_client_helper=self._catalog_client_helper,
            hive_client=self._hive_client,
        )
        self._training_scoring_client = TrainingScoringClient(
            catalog_client=self._catalog_client,
            catalog_client_helper=self._catalog_client_helper,
            hive_client=self._hive_client,
            model_registry_uri=model_registry_uri,
        )
        self._publish_client = PublishClient(
            catalog_client=self._catalog_client,
            hive_client=self._hive_client,
            hive_client_helper=self._hive_client_helper,
        )
        self._model_registry_uri = model_registry_uri

    @deprecated(
        "FeatureStoreClient.create_table",
        since="v0.3.6",
    )
    def create_feature_table(
        self,
        name: str,
        keys: Union[str, List[str]],
        features_df: DataFrame = None,
        schema: StructType = None,
        partition_columns: Union[str, List[str]] = None,
        description: str = None,
        timestamp_keys: Union[str, List[str]] = None,
        **kwargs,
    ) -> FeatureTable:
        path = kwargs.pop("path", None)
        validation_utils.check_kwargs_empty(kwargs, "create_feature_table")

        return self._compute_client._create_table(
            name,
            keys,
            features_df,
            schema=schema,
            partition_columns=partition_columns,
            description=description,
            timestamp_keys=timestamp_keys,
            path=path,
            req_context=RequestContext(request_context.CREATE_FEATURE_TABLE),
        )

    # TODO [ML-15539]: Replace the bitly URL in doc string
    def create_table(
        self,
        name: str,
        primary_keys: Union[str, List[str]],
        df: Optional[DataFrame] = None,
        *,
        timestamp_keys: Union[str, List[str], None] = None,
        partition_columns: Union[str, List[str], None] = None,
        schema: Optional[StructType] = None,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> FeatureTable:
        """
        Create and return a feature table with the given name and primary keys.

        The returned feature table has the given name and primary keys.
        Uses the provided ``schema`` or the inferred schema
        of the provided ``df``. If ``df`` is provided, this data will be saved in
        a Delta table. Supported data types for features are: ``IntegerType``, ``LongType``,
        ``FloatType``, ``DoubleType``, ``StringType``, ``BooleanType``, ``DateType``,
        ``TimestampType``, ``ShortType``, ``ArrayType``, ``MapType``, and ``BinaryType``,
        and ``DecimalType``.

        :param name: A feature table name of the form ``<database_name>.<table_name>``,
          for example ``dev.user_features``.
        :param primary_keys: The feature table's primary keys. If multiple columns are required,
          specify a list of column names, for example ``['customer_id', 'region']``.
        :param df: Data to insert into this feature table. The schema of
          ``df`` will be used as the feature table schema.
        :param timestamp_keys: Columns containing the event time associated with feature value.
          Timestamp keys and primary keys of the feature table uniquely identify the feature value
          for an entity at a point in time.


          .. note::

             Experimental: This argument may change or be removed in
             a future release without warning.

        :param partition_columns: Columns used to partition the feature table. If a list is
          provided, column ordering in the list will be used for partitioning.

          .. Note:: When choosing partition columns for your feature table, use columns that do
                    not have a high cardinality. An ideal strategy would be such that you
                    expect data in each partition to be at least 1 GB.
                    The most commonly used partition column is a ``date``.

                    Additional info: `Choosing the right partition columns for Delta tables
                    <https://bit.ly/3ueXsjv>`_
        :param schema: Feature table schema. Either ``schema`` or ``df`` must be provided.
        :param description: Description of the feature table.
        :param tags: Tags to associate with the feature table.

          .. note::

            Available in version >= 0.4.1.

        :Other Parameters:
          * **path** (``Optional[str]``) --
            Path in a supported filesystem. Defaults to the database location.
        """

        return self._compute_client.create_table(
            name=name,
            primary_keys=primary_keys,
            df=df,
            timestamp_keys=timestamp_keys,
            partition_columns=partition_columns,
            schema=schema,
            description=description,
            tags=tags,
            **kwargs,
        )

    def register_table(
        self,
        *,
        delta_table: str,
        primary_keys: Union[str, List[str]],
        timestamp_keys: Union[str, List[str], None] = None,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
    ) -> FeatureTable:
        """
        Register an existing Delta table as a feature table with the given primary keys.

        The returned feature table has the same name as the Delta table.

        .. note::

          Available in version >= 0.3.8.

        :param name: A Delta table name of the form ``<database_name>.<table_name>``,
          for example ``dev.user_features``. The table must exist in the metastore.
        :param primary_keys: The Delta table's primary keys. If multiple columns are required,
          specify a list of column names, for example ``['customer_id', 'region']``.
        :param timestamp_keys: Columns containing the event time associated with feature value.
          Together, the timestamp keys and primary keys uniquely identify the feature value at a point in time.
        :param description: Description of the feature table.
        :param tags: Tags to associate with the feature table.

          .. note::

            Available in version >= 0.4.1.

        :return: A :class:`FeatureTable <databricks.feature_store.entities.feature_table.FeatureTable>` object.
        """

        return self._compute_client.register_table(
            delta_table=delta_table,
            primary_keys=primary_keys,
            timestamp_keys=timestamp_keys,
            description=description,
            tags=tags,
        )

    def get_table(self, name: str) -> FeatureTable:
        """
        Get a feature table's metadata.

        :param name: A feature table name of the form ``<database_name>.<table_name>``, for
          example ``dev.user_features``.
        """
        return self._compute_client.get_table(
            name=name, req_context=RequestContext(request_context.GET_TABLE)
        )

    @experimental
    def drop_table(self, name: str) -> None:
        """
        Delete the specified feature table. This API also drops the underlying Delta table.

        .. note::

          Available in version >= 0.4.1.

        :param name: The feature table name of the form ``<database_name>.<table_name>``,
            for example ``dev.user_features``.

        .. note::
            Deleting a feature table can lead to unexpected failures in  upstream producers and
            downstream consumers (models, endpoints, and scheduled jobs). You must delete any existing
            published online stores separately.
        """
        self._compute_client.drop_table(name=name)

    @deprecated(
        "FeatureStoreClient.get_table",
        since="v0.3.6",
    )
    def get_feature_table(self, name: str) -> FeatureTable:
        return self._compute_client.get_table(
            name=name, req_context=RequestContext(request_context.GET_FEATURE_TABLE)
        )

    def read_table(self, name: str, **kwargs) -> DataFrame:
        """
        Read the contents of a feature table.

        :param name: A feature table name of the form ``<database_name>.<table_name>``, for
          example ``dev.user_features``.
        :return: The feature table contents, or an exception will be raised if this feature table does not
          exist.
        """
        return self._compute_client.read_table(name=name, **kwargs)

    def _get_feature_names_for_tables(
        self, req_context: RequestContext, table_names: Set[str]
    ) -> Dict[str, List[Feature]]:
        """
        Lookup features from the feature catalog for all table_names, return a dictionary of tablename -> list of features.
        """
        return {
            table_name: self._catalog_client.get_features(table_name, req_context)
            for table_name in table_names
        }

    def _get_feature_table_metadata_for_tables(
        self, req_context: RequestContext, table_names: Set[str]
    ) -> Dict[str, FeatureTable]:
        """
        Lookup FeatureTable metadata from the feature catalog for all table_names, return a dictionary of tablename -> FeatureTable.
        """
        return {
            table_name: self._catalog_client.get_feature_table(table_name, req_context)
            for table_name in table_names
        }

    def _load_feature_data_for_tables(
        self, table_names: Set[str]
    ) -> Dict[str, DataFrame]:
        """
        Load feature DataFrame objects for all table_names, return a dictionary of tablename -> DataFrame.
        """
        return {
            table_name: self._hive_client.read_table(table_name)
            for table_name in table_names
        }

    def _explode_feature_lookups(
        self,
        feature_lookups: List[FeatureLookup],
        feature_table_features_map: Dict[str, List[Feature]],
        feature_table_metadata_map: Dict[str, FeatureTable],
    ) -> List[FeatureColumnInfo]:
        """
        Explode FeatureLookups and collect into FeatureColumnInfos.  A FeatureLookup may explode into either:

        1. A single FeatureColumnInfo, in the case where only a single feature name is specified.
        2. Multiple FeatureColumnInfos, in the cases where either multiple or all feature names are specified.

        Additionally, when all feature names are specified in a FeatureLookup via setting feature_names to None,
        FeatureColumnInfos will be created for all features except primary keys.
        The order of the FeatureColumnInfos returned by this method will be the same order as returned by
        the backend:

        - All partition keys that are not primary keys, in the partition key order
        - All other non-key features
        """
        feature_column_infos = []
        for feature_lookup in feature_lookups:
            feature_column_infos_for_feature_lookup = self._explode_feature_lookup(
                feature_lookup=feature_lookup,
                features=feature_table_features_map[feature_lookup.table_name],
                feature_table=feature_table_metadata_map[feature_lookup.table_name],
            )
            feature_column_infos += feature_column_infos_for_feature_lookup
        return feature_column_infos

    def _explode_feature_lookup(
        self,
        feature_lookup: FeatureLookup,
        features: List[Feature],
        feature_table: FeatureTable,
    ) -> List[FeatureColumnInfo]:
        feature_names = []
        if feature_lookup._get_feature_names():
            # If the user explicitly passed in a feature name or list of feature names, use that
            feature_names += feature_lookup._get_feature_names()
        else:
            # Otherwise assume the user wants all columns in the feature table
            feature_names += [
                feature.name
                for feature in features
                # Filter out primary keys and timestamp keys
                if (
                    feature.name
                    not in [*feature_table.primary_keys, *feature_table.timestamp_keys]
                )
            ]

        return [
            FeatureColumnInfo(
                table_name=feature_lookup.table_name,
                feature_name=feature_name,
                lookup_key=utils.as_list(feature_lookup.lookup_key),
                output_name=(feature_lookup._get_output_name(feature_name)),
                timestamp_lookup_key=utils.as_list(
                    feature_lookup.timestamp_lookup_key, default=[]
                ),
            )
            for feature_name in feature_names
        ]

    def write_table(
        self,
        name: str,
        df: DataFrame,
        mode: str = MERGE,
        checkpoint_location: Optional[str] = None,
        trigger: Dict[str, Any] = _DEFAULT_WRITE_STREAM_TRIGGER,
    ) -> Optional[StreamingQuery]:
        """
        Writes to a feature table.

        If the input :class:`DataFrame <pyspark.sql.DataFrame>` is streaming, will create a write stream.

        :param name: A feature table name of the form ``<database_name>.<table_name>``,
          for example ``dev.user_features``. Raises an exception if this feature table does not
          exist.
        :param df: Spark :class:`DataFrame <pyspark.sql.DataFrame>` with feature data. Raises an exception if the schema does not
          match that of the feature table.
        :param mode: Two supported write modes:

          * ``"overwrite"`` updates the whole table.

          * ``"merge"`` will upsert the rows in ``df`` into the feature table. If ``df`` contains
            columns not present in the feature table, these columns will be added as new features.

        :param checkpoint_location: Sets the Structured Streaming ``checkpointLocation`` option.
          By setting a ``checkpoint_location``, Spark Structured Streaming will store
          progress information and intermediate state, enabling recovery after failures.
          This parameter is only supported when the argument ``df`` is a streaming :class:`DataFrame <pyspark.sql.DataFrame>`.
        :param trigger: If ``df.isStreaming``, ``trigger`` defines the timing of stream data
          processing, the dictionary will be unpacked and passed to :meth:`DataStreamWriter.trigger <pyspark.sql.streaming.DataStreamWriter.trigger>`
          as arguments. For example, ``trigger={'once': True}`` will result in a call to
          ``DataStreamWriter.trigger(once=True)``.
        :return: If ``df.isStreaming``, returns a PySpark :class:`StreamingQuery <pyspark.sql.streaming.StreamingQuery>`. :obj:`None` otherwise.
        """
        return self._compute_client.write_table(
            name=name,
            df=df,
            mode=mode,
            req_context=RequestContext(request_context.WRITE_TABLE),
            checkpoint_location=checkpoint_location,
            trigger=trigger,
            producer_action=ProducerAction.WRITE,
        )

    @staticmethod
    def _index_of_online_store(
        online_store: OnlineStoreSpec, online_stores_list: List[Dict[str, str]]
    ):
        """
        TODO (ML-21692): Do not use this function until ML-21692 is resolved.

        Get the index of the online store.

        Returns the index of the provided online_store in the online_stores_list, or
        None.

        Two online stores are considered the same if they have the same
        database_name, host, and port.

        :param online_store: If the online_store is in the online_stores_list, its
          index will be returned.
        :param online_stores_list: A list of dictionaries of the properties of each
          online store.
        """

        def store_identifier(store_properties):
            return (
                store_properties[DATABASE_NAME],
                store_properties[TABLE_NAME],
                store_properties[HOSTNAME],
                store_properties[PORT],
            )

        online_store_identifier = store_identifier(online_store.non_secret_properties())
        online_store_to_index = {
            store_identifier(store): i for i, store in enumerate(online_stores_list)
        }
        return online_store_to_index.get(online_store_identifier)

    @experimental
    def add_data_sources(
        self,
        *,
        feature_table_name: str,
        source_names: Union[str, List[str]],
        source_type: str = "custom",
    ) -> None:
        """
        Add data sources to the feature table.

        :param feature_table_name: The feature table name.
        :param source_names: Data source names. For multiple sources,
            specify a list. If a data source name already exists, it is ignored.
        :param source_type: One of the following:

            * ``"table"``: Table in format <database_name>.<table_name> and is stored in the metastore (eg Hive).

            * ``"path"``: Path, eg in the Databricks File System (DBFS).

            * ``"custom"``: Manually added data source, neither a table nor a path.
        """
        req_context = RequestContext(request_context.ADD_DATA_SOURCES)
        source_type_string = source_type.strip().lower()
        if source_type_string not in self._DATA_SOURCE_TYPES:
            supported_source_type_list = ", ".join(
                [f"'{m}'" for m in self._DATA_SOURCE_TYPES]
            )
            raise ValueError(
                f"Unsupported source_type '{source_type_string}'. Use one of ({supported_source_type_list})"
            )
        validation_utils.check_qualified_name(feature_table_name)
        feature_table = self._catalog_client.get_feature_table(
            feature_table_name, req_context
        )
        if source_type_string == PATH:
            existing_sources_of_same_type = feature_table.path_data_sources
        elif source_type_string == TABLE:
            existing_sources_of_same_type = feature_table.table_data_sources
        elif source_type_string == CUSTOM:
            existing_sources_of_same_type = feature_table.custom_data_sources

        source_names_as_list = utils.as_list(source_names)
        duplicate_sources = [
            source
            for source in source_names_as_list
            if source in existing_sources_of_same_type
        ]
        if duplicate_sources:
            _logger.info(
                f"The following data source of type {source_type_string} already exists and will not be added: {duplicate_sources}."
            )

        new_sources = [
            source
            for source in source_names_as_list
            if source not in existing_sources_of_same_type
        ]
        if new_sources:
            tables = set(new_sources) if source_type_string == TABLE else set()
            paths = set(new_sources) if source_type_string == PATH else set()
            custom_sources = set(new_sources) if source_type_string == CUSTOM else set()
            self._catalog_client.add_data_sources(
                feature_table_name,
                tables=list(tables),
                paths=list(paths),
                custom_sources=list(custom_sources),
                req_context=req_context,
            )

    @experimental
    def delete_data_sources(
        self,
        *,
        feature_table_name: str,
        source_names: Union[str, List[str]],
    ) -> None:
        """
        Delete data sources from the feature table.

        .. Note:: Data sources of all types (table, path, custom) that match the source names will be deleted.

        :param feature_table_name: The feature table name.
        :param source_names: Data source names. For multiple sources,
            specify a list. If a data source name does not exist,
            it is ignored.

        """
        req_context = RequestContext(request_context.DELETE_DATA_SOURCES)
        validation_utils.check_qualified_name(feature_table_name)
        feature_table = self._catalog_client.get_feature_table(
            feature_table_name, req_context
        )
        existing_data_sources = (
            feature_table.path_data_sources
            + feature_table.table_data_sources
            + feature_table.custom_data_sources
        )
        source_names_as_list = utils.as_list(source_names)
        # initialize as a set to remove duplicate input source
        existing_data_sources_to_delete = {
            source for source in source_names_as_list if source in existing_data_sources
        }
        invalid_data_sources = [
            source
            for source in source_names_as_list
            if source not in existing_data_sources
        ]
        if invalid_data_sources:
            _logger.info(
                f"The following data sources do not exist and will not be deleted: {invalid_data_sources}."
            )
        if existing_data_sources_to_delete:
            self._catalog_client.delete_data_sources(
                feature_table_name,
                list(existing_data_sources_to_delete),
                req_context,
            )

    def publish_table(
        self,
        name: str,
        online_store: OnlineStoreSpec,
        *,
        filter_condition: Optional[str] = None,
        mode: str = MERGE,
        streaming: bool = False,
        checkpoint_location: Optional[str] = None,
        trigger: Dict[str, Any] = _DEFAULT_PUBLISH_STREAM_TRIGGER,
        features: Union[str, List[str], None] = None,
    ) -> Optional[StreamingQuery]:
        """
        Publish a feature table to an online store.

        :param name: Name of the feature table.
        :param online_store: Specification of the online store.
        :param filter_condition: A SQL expression using feature table columns that filters feature
          rows prior to publishing to the online store. For example, ``"dt > '2020-09-10'"``. This
          is analogous to running ``df.filter`` or a ``WHERE`` condition in SQL on a feature table
          prior to publishing.
        :param mode: Specifies the behavior when data already exists in this feature
          table in the online store. If ``"overwrite"`` mode is used, existing data is
          replaced by the new data. If ``"merge"`` mode is used, the new data will be
          merged in, under these conditions:

          * If a key exists in the online table but not the offline table,
            the row in the online table is unmodified.

          * If a key exists in the offline table but not the online table,
            the offline table row is inserted into the online table.

          * If a key exists in both the offline and the online tables,
            the online table row will be updated.

        :param streaming: If ``True``, streams data to the online store.
        :param checkpoint_location: Sets the Structured Streaming ``checkpointLocation`` option.
          By setting a ``checkpoint_location``, Spark Structured Streaming will store
          progress information and intermediate state, enabling recovery after failures.
          This parameter is only supported when ``streaming=True``.
        :param trigger: If ``streaming=True``, ``trigger`` defines the timing of
          stream data processing. The dictionary will be unpacked and passed
          to :meth:`DataStreamWriter.trigger <pyspark.sql.streaming.DataStreamWriter.trigger>` as arguments. For example, ``trigger={'once': True}``
          will result in a call to ``DataStreamWriter.trigger(once=True)``.
        :param features: Specifies the feature column(s) to be published to the online store.
          The selected features must be a superset of existing online store features. Primary key columns
          and timestamp key columns will always be published.

          .. Note:: This parameter is only supported when ``mode="merge"``. When ``features`` is not set, the whole feature table will be published.

        :return: If ``streaming=True``, returns a PySpark :class:`StreamingQuery <pyspark.sql.streaming.StreamingQuery>`, :obj:`None` otherwise.
        """
        return self._publish_client._publish_table(
            name=name,
            online_store=online_store,
            filter_condition=filter_condition,
            mode=mode,
            streaming=streaming,
            checkpoint_location=checkpoint_location,
            trigger=trigger,
            features=features,
        )

    def create_training_set(
        self,
        df: DataFrame,
        feature_lookups: List[FeatureLookup],
        label: Union[str, List[str], None],
        exclude_columns: List[str] = [],
    ) -> TrainingSet:
        """
        Create a :class:`TrainingSet <databricks.feature_store.training_set.TrainingSet>`.

        :param df: The :class:`DataFrame <pyspark.sql.DataFrame>` used to join features into.
        :param feature_lookups: List of features to join into the :class:`DataFrame <pyspark.sql.DataFrame>`.
        :param label: Names of column(s) in :class:`DataFrame <pyspark.sql.DataFrame>` that contain training set labels. To create a training set without a label field, i.e. for unsupervised training set, specify label = None.
        :param exclude_columns: Names of the columns to drop from the :class:`TrainingSet <databricks.feature_store.training_set.TrainingSet>` :class:`DataFrame <pyspark.sql.DataFrame>`.
        :return: A :class:`TrainingSet <databricks.feature_store.training_set.TrainingSet>` object.
        """

        return self._training_scoring_client.create_training_set(
            df=df,
            feature_lookups=feature_lookups,
            label=label,
            exclude_columns=exclude_columns,
        )

    def log_model(
        self,
        model: Any,
        artifact_path: str,
        *,
        flavor: ModuleType,
        training_set: TrainingSet,
        registered_model_name: Optional[str] = None,
        await_registration_for: int = mlflow.tracking._model_registry.DEFAULT_AWAIT_MAX_SLEEP_SECONDS,
        **kwargs,
    ):
        """
        Log an MLflow model packaged with feature lookup information.

        .. note::

           The :class:`DataFrame <pyspark.sql.DataFrame>` returned
           by :meth:`.TrainingSet.load_df` **must** be used to train the
           model. If it has been modified (for example data normalization, add a column,
           and similar), these modifications will not be applied at inference time,
           leading to training-serving skew.

        .. todo::

           [ML-15539]: Replace the bitly URL in doc string


        :param model: Model to be saved. This model must be capable of being saved by
          ``flavor.save_model``. See the `MLflow Model API
          <https://bit.ly/3yzl1r0>`_.
        :param artifact_path: Run-relative artifact path.
        :param flavor: MLflow module to use to log the model. ``flavor`` should have
          type :obj:`ModuleType <types.ModuleType>`.
          The module must have a method ``save_model``, and must support the ``python_function``
          flavor. For example, :mod:`mlflow.sklearn`, :mod:`mlflow.xgboost`, and similar.
        :param training_set: The :class:`.TrainingSet` used to train this model.
        :param registered_model_name:

          .. note::

             Experimental: This argument may change or be removed in
             a future release without warning.

          If given, create a model version under ``registered_model_name``,
          also creating a registered model if one with the given name does not exist.
        :param await_registration_for: Number of seconds to wait for the model version to finish
          being created and is in ``READY`` status. By default, the function waits for five minutes.
          Specify ``0`` or :obj:`None` to skip waiting.
        :return: `None`
        """
        self._training_scoring_client.log_model(
            model=model,
            artifact_path=artifact_path,
            flavor=flavor,
            training_set=training_set,
            registered_model_name=registered_model_name,
            await_registration_for=await_registration_for,
            **kwargs,
        )

    def score_batch(
        self, model_uri: str, df: DataFrame, result_type: str = "double"
    ) -> DataFrame:
        """
        Evaluate the model on the provided :class:`DataFrame <pyspark.sql.DataFrame>`.

        Additional features required for
        model evaluation will be automatically retrieved from :mod:`Feature Store <databricks.feature_store.client>`.

        .. todo::

           [ML-15539]: Replace the bitly URL in doc string

        The model must have been logged with :meth:`.FeatureStoreClient.log_model`,
        which packages the model with feature metadata. Unless present in ``df``,
        these features will be looked up from :mod:`Feature Store <databricks.feature_store.client>` and joined with ``df``
        prior to scoring the model.

        If a feature is included in ``df``, the provided feature values will be used rather
        than those stored in :mod:`Feature Store <databricks.feature_store.client>`.

        For example, if a model is trained on two features ``account_creation_date`` and
        ``num_lifetime_purchases``, as in:

        .. code-block:: python

            feature_lookups = [
                FeatureLookup(
                    table_name = 'trust_and_safety.customer_features',
                    feature_name = 'account_creation_date',
                    lookup_key = 'customer_id',
                ),
                FeatureLookup(
                    table_name = 'trust_and_safety.customer_features',
                    feature_name = 'num_lifetime_purchases',
                    lookup_key = 'customer_id'
                ),
            ]

            with mlflow.start_run():
                training_set = fs.create_training_set(
                    df,
                    feature_lookups = feature_lookups,
                    label = 'is_banned',
                    exclude_columns = ['customer_id']
                )
                ...
                  fs.log_model(
                    model,
                    "model",
                    flavor=mlflow.sklearn,
                    training_set=training_set,
                    registered_model_name="example_model"
                  )

        Then at inference time, the caller of :meth:`FeatureStoreClient.score_batch` must pass
        a :class:`DataFrame <pyspark.sql.DataFrame>` that includes ``customer_id``, the ``lookup_key`` specified in the
        ``FeatureLookups`` of the :mod:`training_set <databricks.feature_store.training_set>`.
        If the :class:`DataFrame <pyspark.sql.DataFrame>` contains a column
        ``account_creation_date``, the values of this column will be used
        in lieu of those in :mod:`Feature Store <databricks.feature_store.client>`. As in:

        .. code-block:: python

            # batch_df has columns ['customer_id', 'account_creation_date']
            predictions = fs.score_batch(
                'models:/example_model/1',
                batch_df
            )

        :param model_uri: The location, in URI format, of the MLflow model logged using
          :meth:`FeatureStoreClient.log_model`. One of:

            * ``runs:/<mlflow_run_id>/run-relative/path/to/model``

            * ``models:/<model_name>/<model_version>``

            * ``models:/<model_name>/<stage>``

          For more information about URI schemes, see
          `Referencing Artifacts <https://bit.ly/3wnrseE>`_.
        :param df: The :class:`DataFrame <pyspark.sql.DataFrame>` to score the model on. :mod:`Feature Store <databricks.feature_store.client>` features will be joined with
          ``df`` prior to scoring the model. ``df`` must:

              1. Contain columns for lookup keys required to join feature data from Feature
              Store, as specified in the ``feature_spec.yaml`` artifact.

              2. Contain columns for all source keys required to score the model, as specified in
              the ``feature_spec.yaml`` artifact.

              3. Not contain a column ``prediction``, which is reserved for the model's predictions.
              ``df`` may contain additional columns.

        :param result_type: The return type of the model.
           See :func:`mlflow.pyfunc.spark_udf` result_type.
        :return: A :class:`DataFrame <pyspark.sql.DataFrame>`
           containing:

            1. All columns of ``df``.

            2. All feature values retrieved from Feature Store.

            3. A column ``prediction`` containing the output of the model.

        """
        return self._training_scoring_client.score_batch(
            model_uri=model_uri, df=df, result_type=result_type
        )

    def set_feature_table_tag(self, *, table_name: str, key: str, value: str) -> None:
        """Create or update a tag associated with the feature table. If the tag with the
        corresponding key already exists, its value will be overwritten with the new value.

        .. note::

          Available in version >= 0.4.1.

        :param table_name: the feature table name
        :param key: tag key
        :param value: tag value
        """
        self._compute_client.set_feature_table_tag(
            table_name=table_name, key=key, value=value
        )

    def delete_feature_table_tag(self, *, table_name: str, key: str) -> None:
        """Delete the tag associated with the feature table. Deleting a non-existent tag will emit a warning.

        .. note::

          Available in version >= 0.4.1.

        :param table_name: the feature table name.
        :param key: the tag key to delete.
        """
        self._compute_client.delete_feature_table_tag(table_name=table_name, key=key)
